package com.owlfish.cborrpc;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectReader;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.PropertyNamingStrategy.PascalCaseStrategy;
import com.fasterxml.jackson.databind.exc.UnrecognizedPropertyException;
import com.fasterxml.jackson.databind.SequenceWriter;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.dataformat.cbor.*;
import java.io.*;
import java.net.*;
import java.util.HashMap;

/**
 * CBORRpcClient is a multi-thread safe RPC client for CBOR based RPC implementations using the Golang RPC protocol.
 * <p>
 * The RPC protocol consists of sending two objects to make an RPC call:
 * <ol>
 * <li>A header object: {"ServiceMethod": "service.method", "Seq": sequenceNumber} where the service.method is the name of the service being called 
 * and sequenceNumber is a 64 bit number that is unique to this TCP session.</li>
 * <li>The parameters object required by the service.</li>
 * </ol>
 * <p>
 * The RPC response also consists of two objects:
 * <ol>
 * <li>A response object: {"ServiceMethod": "service.method", "Seq": sequenceNumber, "Error": errorMsg}.  The sequenceNumber is used to relate this
 * RPC response to the call that was made and allows for multiple calls to be interleaved and returned in any order.  The Error field is an empty
 * string for a successful call, or the error message if one occured.</li>
 * <li>The response object returned by the service.  This will be sent even if Error is not empty, although values in the response object may not
 * have been populated depending on the error.</li>
 * </ol>
*/
public class CBORRpcClient implements Runnable {
	private String hostname;
	private int port;
	private Socket connection;
	private BufferedInputStream input;
	private BufferedOutputStream output;
	private ObjectMapper mapper;
	private CBORFactory factory;
	private CBORGenerator generator;
	private int sequence;
	private HashMap<Integer,RpcResult> calls;
	private Thread readingThread;

	/**
	 * Creates an instance of the CBORRpcClient connected to the given hostname and port.
	 * @param hostname The hostname to connect to.
	 * @param port The post number to connect to.
	 * @throws IOException Thrown if there is an error connecting to the server. 
	*/
	public CBORRpcClient (String hostname, int port) throws IOException {
		this.hostname = hostname;
		this.port = port;
		connection = new Socket (hostname, port);
		input = new BufferedInputStream(this.connection.getInputStream());
		output = new BufferedOutputStream(this.connection.getOutputStream());
		factory = new CBORFactory();
		factory.configure (JsonParser.Feature.AUTO_CLOSE_SOURCE, false);
		mapper = new ObjectMapper(factory);
		mapper = mapper.configure (JsonParser.Feature.AUTO_CLOSE_SOURCE, false);
		mapper = mapper.configure (JsonGenerator.Feature.AUTO_CLOSE_TARGET, false);
		mapper.setPropertyNamingStrategy(new PascalCaseStrategy());
		this.generator = factory.createGenerator(output);
		generator.configure (JsonGenerator.Feature.AUTO_CLOSE_TARGET, false);
		generator.configure (JsonGenerator.Feature.FLUSH_PASSED_TO_STREAM, false);
		calls = new HashMap<Integer,RpcResult>();
		sequence = 1;
		readingThread = new Thread (this);
		readingThread.start();

	}

	/**
	 * Calls an RPC method on the server that this client is connected to.  The call will block until either a response is received or an
	 * exception occurs.
	 * @param method the name of the method to call, usually in "object.method" format.
	 * @param argument the object that will be serialised as the argument to the RPC method.  Can be a Map&lt;String,Object&gt; or a POJO.
	 * @param resultClass the Class of the object that the response will be deserialised into.  Can be a Map&lt;String,Object&gt; or a POJO.
	 * @return object of class resultClass that was returned from the RPC server.
	 * @throws RPCException is thrown if an IO Exception occurs, a deserialisation error occurs or the connection is lost.
	*/
	public Object Call (String method, Object argument, Class<?> resultClass) throws RPCException {
		// Start by registering a call response
		RpcResult pendingResult = new RpcResult (resultClass);
		Integer key;
		synchronized (this.calls) {
			key = Integer.valueOf (this.sequence++);
			this.calls.put (key, pendingResult);
		}

		// Now lock the output stream and write out the values.
		synchronized (this.output) {
			try {
				// Write out the method RPC header
				generator.writeStartObject();
				generator.writeStringField ("ServiceMethod", method);
				generator.writeFieldName ("Seq");
				generator.writeNumber (key.intValue());
				generator.writeEndObject();

				if (argument == null) {
					generator.writeStartObject();
					generator.writeEndObject();
					generator.flush();
				} else {
					generator.flush();
					//System.out.println("Writing argument");
					mapper.writeValue(output, argument);
					//System.out.println("Done writign.");
				}
				//System.out.println ("Resulting CBOR: " + arrayout.toString());

				output.flush();
			}
			catch (IOException e) {
				// Close the socket and return an exception.
			}
		}
		// We've released the lock - wait for a response
		return pendingResult.WaitForResponse();
	}

	/**
	 * Closes the connection to the RPC server.
	 * <p>
	 * This will force any oustanding RPC calls to throw RPCExceptions and renders the CBORRpcClient unusable.
	*/
	public void close() {
		//System.out.println("Closing socket connection");
		try {
			this.connection.close();
		} catch (IOException e) {

		}
	}

	/**
	 * A thread created by the constructor runs here to handle responses coming back from the RPCServer.
	*/
	public void run() {
		boolean running = true;
		RpcResult result = null;
		Object returnObj = null;

		JsonParser parser = null;
		ObjectReader reader = null;
		try {
			parser = factory.createParser(input);
			parser.configure (JsonParser.Feature.AUTO_CLOSE_SOURCE, false);
			reader = mapper.reader();
		}
		catch (IOException ioe) {
			running = false;
		}

		while (running) {
			try {
				Response response = reader.readValue (parser, Response.class);

				// Lock the hashmap while we find the result object
				Integer key = Integer.valueOf (response.Seq);
				synchronized (this.calls) {
					result = this.calls.get (key);
					this.calls.remove (key);
				}
				if (result != null) {
					//System.out.println("Reading response");
					RPCException returnException = null;
					try {
						returnObj = reader.readValue (parser, result.klass);
					}
					catch (com.fasterxml.jackson.databind.exc.UnrecognizedPropertyException e) {
						// Bad RPC excpetion.
						returnException = new RPCProtocolException ("Exception parsing RPC response: " + e.toString());
					}
					if (! response.Error.equals ("")) {
						// Create an exception to be thrown in the sending thread.
						returnException = new RPCErrorException (response.Error);
					}
					// Tell the waiting send thread what the results are.
					result.SendResponse(returnObj, returnException);
				} else {
					throw new RPCProtocolException ("Unable to find response object for RPC result.");
				}
			} catch (Exception e) {
				running = false;
			}
		}
		// If we get there - clear all waiting clients with an exception.
		try {
			this.connection.close();
		} catch (IOException ioe) {}
		synchronized (this.calls) {
			for (RpcResult outstanding: this.calls.values()) {
				outstanding.SendResponse(null, new RPCIOException ("Connection lost before RPC call returned."));
			}
		}
	}
}