package com.owlfish.cborrpc;

/**
 * RpcResult represents a pending RPC call and encapsulates the results of the call.  It is used internally within the RPC client.
*/
class RpcResult {
	private RPCException exception;
	public Class<?> klass;
	private Object result;
	private boolean received;

	/**
	 * Creates a new pending result with the given type to represent the result.
	 * @param klass The Class of the object that the RPC result should be deserialised into.
	*/
	public RpcResult (Class<?> klass) {
		this.received = false;
		this.klass = klass;
	}

	/**
	 * WaitForResponse blocks in the RPC calling thread waiting on the response.  This allows multiple threads to call RPC methods interleaved.
	 * @return Object The deserialised result - an instance of klass.
	 * @throws RPCException This will be thrown if an error is returned from the RPC method or an IO exception occurs. 
	*/
	public synchronized Object WaitForResponse () throws RPCException {
		while (! this.received) {
			try {
				this.wait();
			} catch(InterruptedException e){
        	}
		}
		if (this.exception != null) {
			throw this.exception;
		}
		return this.result;
	}

	/**
	 * SendResponse is used by the RPC receive thread to notify the caller that a response is available.
	 * @param result The resulting object from the RPC call, an instance of klass.
	 * @param errorException Any exception that may have happened in the receive, e.g. Error being returned from the RPC call.
	*/
	public synchronized void SendResponse (Object result, RPCException errorException) {
		this.result = result;
		this.exception = errorException;
		this.received = true;
		this.notify();
	}
}
