package com.owlfish.forestbus;

import com.owlfish.cborrpc.*;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
* The ForestClient provides a multi-thread safe client for interfacing to a Forest Bus cluster of nodes.  The client handles establishing
* the connectivity to the individual nodes and finding the current location of leaders for given topics.
* <p>
* ForestClient connects using CBOR RPC.
*/
public class ForestClient {
	private String clusterID;
	private String[] nodes;
	private HashMap<String,CBORRpcClient> connections;
	private HashMap<String,String> topicLeaders;

	/**
	* Creates a ForestClient ready for use.  Connections to nodes happens on demand rather than in the constructor.
	*
	* @param clusterID the id the cluster is configured with.  If this is incorrect ClusterIDMismatchException exceptions will be thrown.
	* @param nodes the array of nodes to connect to.  The strings must be in the format name:port and match those given on the nodes when started up with -cbor.
	*/
	public ForestClient (String clusterID, String[] nodes) {
		this.clusterID = clusterID;
		this.nodes = nodes;
		this.connections = new HashMap<String,CBORRpcClient>();
		this.topicLeaders = new HashMap<String,String>();
	}

	/**
	* getMessages takes the topic, the index, target quantity and a wait flag (boolean) and returns messages from the cluster.
	* <p>
	* GetMessages will usually return more or fewer messages than the quantity requested.  This ensures efficient message retrieval from the node
	* as messages are aligned to offset and cache boundaries.  If any messages are available at the requested index then at least one message 
	* will be returned.
	* <p>
	* If the index requested is no longer available on this node (i.e. clean-up has removed old data) then zero messages will be returned and 
	* the RecieveMessagesResults.nextID will be the index of the first available message.
	* <p>
	* If the messages returned bring the client up to the end of the available messages, then RecieveMessagesResults.nextID will contain the 
	* index of what will become the next message when it has been sent.  By setting wait to true and passing in the index returned by nextID, 
	* GetMessages will block until at least one new message is available, before returning that message/messages.
	*<p>
	* GetMessages throws Exceptions if no nodes are available, a network error occurs or the topic could not be found.  
	* Even if an exception has been thrown - Client.close() must be called to close down all threads.
	*
	* @param topic the name of the topic to get messages from.
	* @param index the index of the first message to retrieve.
	* @param quantity the target number of messages to retrieve.
	* @param wait whether to wait for messages to become available if there are no messages currently at index.
	* @return GetMessagesResult contains receievedMessages (an array of byte arrays) and nextID which contains the next available ID.
	* @throws ForestBusException  Throws ForestBusException or a subclass of this if there is an error.
	*/
	public GetMessagesResult getMessages (String topic, long index, int quantity, boolean wait) throws ForestBusException {
		connectionFactory factory = new connectionFactory();
		while (true) {
			GetMessagesResult result = null;
			CBORRpcClient client = null;
			try {
				client = factory.getClient();
				// NoNodesAvailableException will be thrown if there are no nodes left to talk with
				ReceiveMessagesArgs args = new ReceiveMessagesArgs();
				args.clusterID = clusterID;
				args.topic = topic;
				args.id = index;
				args.quantity = quantity;
				args.waitForMessages = wait;
				//System.out.println("Calling ReceiveMessages with args " + args.toString());

				result = (GetMessagesResult) client.Call ("RPCHandler.ReceiveMessages", args, GetMessagesResult.class);

				if (result.result.code == 1) {
					throw new ClusterIDMismatchException();
				}
				if (result.result.code == 2) {
					throw new TopicNotFoundException ();
				}
				// Catch all for other errors in the server
				if (result.result.code != 0) {
					throw new ForestBusException(result.result.description);
				}
				//System.out.println("Returning result.");
				return result;
			}
			catch (RPCException rpce) {
				//System.out.println ("Exception handled: " + rpce.toString());
				// Do nothing - loop and try another node.
				removeConnection (client);
			}
			
		}
	}

	/**
	* Returns the maximum index available in the given topic.  
	* <p>
	* If the cluster has been completely shutdown and restarted (rather than a rolling restart of individual nodes) then the commit index 
	* may be zero, in which case the maxAvailableIndex will be zero.  Once a message has been sent to the cluster in this topic the commit 
	* index will be recalculated and the maximum commit index will return as normal.
	* @param topic the name of the topic.
	* @return the maximum index number of a message in this topic.
	* @throws ForestBusException  Throws ForestBusException or a subclass of this if there is an error.
	*/
	public long getTopicMaxAvailableIndex (String topic) throws ForestBusException {
		connectionFactory factory = new connectionFactory();
		CBORRpcClient client = null;
		while (true) {
			GetTopicDetailsResults result = null;
			try {
				client = factory.getClient();
				// NoNodesAvailableException will be thrown if there are no nodes left to talk with
				GetTopicDetailsArgs args = new GetTopicDetailsArgs();
				args.topic = topic;
				//System.out.println("Calling GetTopicDetails with args " + args.toString());

				result = (GetTopicDetailsResults) client.Call ("RPCHandler.GetTopicDetails", args, GetTopicDetailsResults.class);
				return result.commitIndex;
			}
			catch (RPCException rpce) {
				//System.out.println ("Exception handled: " + rpce.toString());
				// Do nothing - loop and try another node.
				removeConnection (client);
			}
			
		}
	}

	/**
	* sendMessages sends a batch of messages to the Forest Bus cluster.
	* <p>
	* Messages are a list of sequences of bytes.  Sending many messages (hundreds) at once gives better through-put than sending individual messages.
	* <p>
	* If waitForCommit is false then SendMessages will return as soon as the message has been saved on the leader node for this topic.  
	* If waitForCommit is true then SendMessages will only return once the messages have been replicated to a majority of the nodes in the cluster
	* and are therefore committed.
	* <p>
	* SendMessages throws Exceptions if no leader node is available, a network error occurs or the topic could not be found.  
	* Even if an exception has been thrown - Client.close() must be called to close down all threads.
	* @param topic the name of the topic.
	* @param messages an array of byte arrays containing the messages to be sent.
	* @param waitForCommit flag indicating whether to wait until the messages have been committed to the cluster before returning.
	* @return an array of IDs for the messages sent.
	* @throws ForestBusException  Throws ForestBusException or a subclass of this if there is an error.
	*/
	public long[] sendMessages (String topic, byte[][] messages, boolean waitForCommit) throws ForestBusException {
		connectionFactory factory = new connectionFactory(topic);
		while (true) {
			SendMessagesResults result = null;
			CBORRpcClient client = null;
			try {
				client = factory.getClient();
				// NoNodesAvailableException will be thrown if there are no nodes left to talk with
				SendMessagesArgs args = new SendMessagesArgs();
				args.clusterID = clusterID;
				args.topic = topic;
				args.sentMessages = messages;
				args.waitForCommit = waitForCommit;
				//System.out.println("Calling SendMessages with args " + args.toString());

				result = (SendMessagesResults) client.Call ("RPCHandler.SendMessages", args, SendMessagesResults.class);
				if (result.result.code == 0) {
					return result.ids;
				}
				//System.out.println("Response code " + result.result.description);
				// Else, try a different node.
			}
			catch (RPCException rpce) {
				//System.out.println ("Exception handled: " + rpce.toString());
				// Do nothing - loop and try another node.
				removeConnection (client);
			}
			
		}
	}

	/**
	* Closes all connections to the cluster.  The Client object may still be used, in which case new connections will be opened as required.
	*/
	public void close () {
		synchronized (this.connections) {
			for (CBORRpcClient connection: this.connections.values()) {
				try {
					connection.close();
				} catch (Exception e) {}
			}
			this.connections.clear();
		}
	}

	private void removeConnection (CBORRpcClient client) {
		synchronized (this.connections){
			for (Map.Entry<String,CBORRpcClient> entry : this.connections.entrySet()) {
				if (entry.getValue() == client) {
					try {
						client.close();
					} catch (Exception e) {}
					this.connections.remove (entry.getKey());
					return;
				}
			}
		}
	}

	class connectionFactory {
		private int index;
		private String[] nodeList;
		private String forTopic = null;
		private String preferredNode = null;

		public connectionFactory () {
			nodeList = new String [nodes.length];
			int postition = 0;
			for (String node: nodes) {
				nodeList[postition++] = node;
			}
		}

		public connectionFactory (String topic) {
			this();
			this.forTopic = topic;
			synchronized (topicLeaders) {
				if (topicLeaders.containsKey (topic)) {
					preferredNode = topicLeaders.get (topic);
					int location = Arrays.asList (this.nodeList).indexOf (preferredNode);
					// Move the preferred node to the front of the list.
					if (location > 0) {
						String temp = this.nodeList[0];
						this.nodeList[0] = this.nodeList[location];
						this.nodeList[location] = temp;
					}
				}
			}
		}

		CBORRpcClient getClient() throws ClusterIDMismatchException, NoNodesAvailableException {
			if (this.index >= this.nodeList.length) {
				// We've gone through all nodes - nothing found.
				throw new NoNodesAvailableException();
			}
			
			while (this.index < this.nodeList.length) {
				String nodeToTry = this.nodeList[this.index++];
				//System.out.println ("Looking for connection to node " + nodeToTry);
				CBORRpcClient client = getRPCClient (nodeToTry);
				if (client != null) {
					if (forTopic != null && ! nodeToTry.equals (preferredNode)) {
						// Upate the topic leader for next time.
						synchronized (topicLeaders) {
							topicLeaders.put (forTopic, nodeToTry);
						}
					}
					return client;
				}
			}
			throw new NoNodesAvailableException();
		}
	}

	private CBORRpcClient getRPCClient (String host) throws ClusterIDMismatchException {
		synchronized (this.connections) {
			// Check to see if we already have one
			if (this.connections.containsKey (host)) {
				return this.connections.get (host);
			}
			try {
				String hostname = host.substring (0,host.indexOf(':'));
				String portnumberAsString = host.substring(host.indexOf(':') + 1, host.length());

				CBORRpcClient client = new CBORRpcClient (hostname, Integer.parseInt (portnumberAsString));

				GetClusterDetailsResults results = (GetClusterDetailsResults) client.Call ("RPCHandler.GetClusterDetails", null, GetClusterDetailsResults.class);
				if (! results.clusterID.equals (clusterID)) {
					throw new ClusterIDMismatchException(this.clusterID,results.clusterID);
				}
				this.connections.put (host, client);
				return client;
			}
			catch (RPCException e) {
				return null;
			}
			catch (java.io.IOException ioe) {
				return null;
			}
		}
	}

}